!!  Copyright (C) 2012 Jihuan Tian <tianjh04@gmail.com> and Ling Zhang
!!  <zhangling06@mails.tsinghua.edu.cn>
!!   
!!  This file is part of SPCNDG.
!!
!!  SPCNDG is free software: you can redistribute it and/or modify
!!  it under the terms of the GNU General Public License as published by
!!  the Free Software Foundation, either version 3 of the License, or
!!  (at your option) any later version.
!!
!!  SPCNDG is distributed in the hope that it will be useful,
!!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!!  GNU General Public License for more details.
!!
!!  You should have received a copy of the GNU General Public License
!!  along with SPCNDG.  If not, see <http://www.gnu.org/licenses/>.

module MathFunc
  ! Module defining mathematical functions and constants
  use Defs
  implicit none

  !!!!!!!!!!!!!!!!!!!!!!!!!!!!
  ! Mathematical precisions
  !!!!!!!!!!!!!!!!!!!!!!!!!!!!
  integer, parameter :: SP = selected_real_kind(4) ! single precision
  integer, parameter :: DP = selected_real_kind(8) ! double precision, which is the internal representation for all float numbers

  !!!!!!!!!!!!!!!!!!!!!!!!!!!
  ! Mathematical constants
  !!!!!!!!!!!!!!!!!!!!!!!!!!!
  real(DP), parameter :: pi = 3.1415926535897932_DP
  real(DP), parameter :: eps = 1.0e-30_DP
  real(DP), parameter :: inf = 1.0e30_DP

  ! Default format string for output numbers
  character(len=3) :: int_fmt = 'I10'
  character(len=9) :: dp_fmt = 'ES26.16E3'
  character(len=8) :: sp_fmt = 'ES18.8E3'

  !!!!!!!!!!!!!!!!!!!!!!!!!!
  ! Interface definitions
  !!!!!!!!!!!!!!!!!!!!!!!!!!

  ! Set as null matrix and vector
  interface SetMatZero
     module procedure SetMatZeroI, SetMatZeroD
  end interface SetMatZero

  interface SetVecZero
     module procedure SetVecZeroI, SetVecZeroD
  end interface SetVecZero

  ! Print matrix and vector to standard output
  interface PrintMat
     module procedure PrintMatI, PrintMatD
  end interface PrintMat

  interface PrintColVec
     module procedure PrintColVecI, PrintColVecD
  end interface PrintColVec

  interface PrintRowVec
     module procedure PrintRowVecI, PrintRowVecD
  end interface PrintRowVec

  ! Print matrix and vector into output file with a specified name
  interface PrintVarMat
     module procedure PrintVarMatI, PrintVarMatD
  end interface PrintVarMat

  interface PrintVarColVec
     module procedure PrintVarColVecI, PrintVarColVecD
  end interface PrintVarColVec

  interface PrintVarRowVec
     module procedure PrintVarRowVecI, PrintVarRowVecD
  end interface PrintVarRowVec

  ! Mathematical functions
  interface jacobi
     module procedure jacobi_sc, jacobi_ve
  end interface jacobi
contains
  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  ! Get format strings for different types of data
  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  function GetIntFmt(int_num)
    integer int_num             ! the number of integers to be printed
    character(len=20) int_num_str
    character(len=20) GetIntFmt

    write (int_num_str,'(I20)'), int_num
    GetIntFmt = '('//trim(adjustl(int_num_str))//int_fmt//')'
  end function GetIntFmt

  function GetDPFmt(dp_num)
    integer dp_num              ! the number of double floats to be printed
    character(len=20) dp_num_str
    character(len=20) GetDPFmt

    write (dp_num_str,'(I20)') dp_num
    GetDPFmt = '('//trim(adjustl(dp_num_str))//dp_fmt//')'
  end function GetDPFmt

  function GetSPFmt(sp_num)
    integer sp_num              ! the number of signle floats to be printed
    character(len=20) sp_num_str
    character(len=20) GetSPFmt

    write (sp_num_str,'(I20)') sp_num
    GetSPFmt = '('//trim(adjustl(sp_num_str))//sp_fmt//')'
  end function GetSPFmt

  !!!!!!!!!!!!!!!!!!!!!!!!!!!!
  ! Print matrix and vector
  !!!!!!!!!!!!!!!!!!!!!!!!!!!!

  ! Print matrix and vector to standard output
  subroutine PrintMatI(file_id, array)
    integer file_id
    integer, dimension(:,:), intent(in) :: array
    integer i

    do i = 1, size(array, 1)
       write (file_id, GetIntFmt(size(array, 2))), array(i,:)
    end do
  end subroutine PrintMatI

  subroutine PrintMatD(file_id, array)
    integer file_id
    real(DP), dimension(:,:), intent(in) :: array
    integer i

    do i = 1, size(array, 1)
       write (file_id, GetDPFmt(size(array, 2))), array(i,:)
    end do
  end subroutine PrintMatD

  subroutine PrintColVecI(file_id, array)
    integer file_id
    integer, dimension(:), intent(in) :: array
    integer i

    do i = 1, size(array)
       write (file_id, GetIntFmt(1)), array(i)
    end do
  end subroutine PrintColVecI

  subroutine PrintColVecD(file_id, array)
    integer file_id
    real(DP), dimension(:), intent(in) :: array
    integer i

    do i = 1, size(array)
       write (file_id, GetDPFmt(1)), array(i)
    end do
  end subroutine PrintColVecD

  subroutine PrintRowVecI(file_id, array, M)
    integer file_id, M
    integer, dimension(M), intent(in) :: array

    write (file_id, GetIntFmt(M)), array
  end subroutine PrintRowVecI

  subroutine PrintRowVecD(file_id, array, M)
    integer file_id, M
    real(DP), dimension(M), intent(in) :: array

    write (file_id, GetDPFmt(M)), array
  end subroutine PrintRowVecD

  ! Print matrix and vector into output file with a specified name
  subroutine PrintVarMatI(var_name, var)
    character(len=*), intent(in) :: var_name
    integer, dimension(:,:), intent(in) :: var

    open(unit=tmp_var_id, file=var_name//".dat")
    call PrintMat(tmp_var_id, var)
    close(tmp_var_id)
  end subroutine PrintVarMatI

  subroutine PrintVarMatD(var_name, var)
    character(len=*), intent(in) :: var_name
    real(DP), dimension(:,:), intent(in) :: var

    open(unit=tmp_var_id, file=var_name//".dat")
    call PrintMat(tmp_var_id, var)
    close(tmp_var_id)
  end subroutine PrintVarMatD

  subroutine PrintVarColVecI(var_name, var, M)
    integer M
    character(len=*), intent(in) :: var_name
    integer, dimension(M), intent(in) :: var

    open(unit=tmp_var_id, file=var_name//".dat")
    call PrintColVec(tmp_var_id, var)
    close(tmp_var_id)
  end subroutine PrintVarColVecI

  subroutine PrintVarColVecD(var_name, var, M)
    integer M
    character(len=*), intent(in) :: var_name
    real(DP), dimension(M), intent(in) :: var

    open(unit=tmp_var_id, file=var_name//".dat")
    call PrintColVec(tmp_var_id, var)
    close(tmp_var_id)
  end subroutine PrintVarColVecD

  subroutine PrintVarRowVecI(var_name, var, M)
    integer M
    character(len=*), intent(in) :: var_name
    integer, dimension(M), intent(in) :: var

    open(unit=tmp_var_id, file=var_name//".dat")
    call PrintRowVec(tmp_var_id, var, M)
    close(tmp_var_id)
  end subroutine PrintVarRowVecI

  subroutine PrintVarRowVecD(var_name, var, M)
    integer M
    character(len=*), intent(in) :: var_name
    real(DP), dimension(M), intent(in) :: var

    open(unit=tmp_var_id, file=var_name//".dat")
    call PrintRowVec(tmp_var_id, var, M)
    close(tmp_var_id)
  end subroutine PrintVarRowVecD

  !!!!!!!!!!!!!!!!!!!!!!!!!!!
  ! Mathematical functions
  !!!!!!!!!!!!!!!!!!!!!!!!!!!

  ! Set null matrix
  subroutine SetMatZeroI(array, M, N)
    integer M, N
    integer, dimension(M,N), intent(out) :: array
    integer i, j

    do i = 1, M
       do j = 1, N
          array(i,j) = 0
       end do
    end do
  end subroutine SetMatZeroI

  subroutine SetMatZeroD(array, M, N)
    integer M, N
    real(DP), dimension(M,N), intent(out) :: array
    integer i, j

    do i = 1, M
       do j = 1, N
          array(i,j) = 0.
       end do
    end do
  end subroutine SetMatZeroD

  ! Set null vector
  subroutine SetVecZeroI(array, M)
    integer M
    integer, dimension(M), intent(out) :: array
    integer i

    do i = 1, M
       array(i) = 0
    end do
  end subroutine SetVecZeroI

  subroutine SetVecZeroD(array, M)
    integer M
    real(DP), dimension(M), intent(out) :: array
    integer i

    do i = 1, M
       array(i) = 0.0_DP
    end do
  end subroutine SetVecZeroD

  ! Kronecker delta function
  function deltaij(i, j)
    integer i, j
    real(DP) deltaij

    if (i == j) then
       deltaij = 1.
    else
       deltaij = 0.
    end if
  end function deltaij

  ! Inner product
  function dotprod(p1, p2, N)
    integer N
    real(DP), dimension(N), intent(in) :: p1, p2
    real(DP) dotprod

    dotprod = sum(p1 * p2)
  end function dotprod

  ! Generate identity matrix
  subroutine ones(A, M)
    ! generate unit matrix of dimension MxM
    real(DP), dimension(M,M), intent(out) :: A
    integer i, M

    call SetMatZeroD(A, M, M)
    do i = 1, M
       A(i,i) = 1.0_DP
    end do
  end subroutine ones

  ! Generate zero matrix
  subroutine zeros(A, M, N)
    integer M, N
    real(DP), dimension(M,N), intent(out) :: A

    call SetMatZeroD(A, M, N)
  end subroutine zeros

  ! Jacobi function
  recursive function jacobi_sc(x, alpha, beta, n) result(ja)
    real(DP), intent(in) :: x, alpha, beta
    integer n
    real(DP) ja

    select case (n)
      case (0)
        ja = sqrt(2**(-alpha - beta - 1) * gamma(alpha + beta + 2) / (gamma(alpha + 1) * gamma(beta + 1)))
      case (1)
        ja = 0.5 * sqrt(2**(-alpha - beta - 1) * gamma(alpha + beta + 2) / (gamma(alpha + 1) * gamma(beta + 1))) * &
             & sqrt((alpha + beta + 3) / (alpha + 1) / (beta + 1)) * ((alpha + beta + 2) * x + alpha - beta)
      case default
        ja = (x * jacobi_sc(x, alpha, beta, n - 1) - &
             & CalcJacobiAn(n - 1, alpha, beta) * jacobi_sc(x, alpha, beta, n - 2) - &
             & CalcJacobiBn(n - 1, alpha, beta) * jacobi_sc(x, alpha, beta, n - 1)) / &
             & CalcJacobiAn(n, alpha, beta)
    end select
  end function jacobi_sc

  recursive function jacobi_ve(x, alpha, beta, n) result(ja)
    real(DP), dimension(:), intent(in) :: x
    real(DP), intent(in) :: alpha, beta
    integer n
    real(DP), dimension(:), allocatable :: ja

    if (.not. allocated(ja)) then
      allocate(ja(size(x)))
    end if

    select case (n)
      case (0)
        ja = sqrt(2**(-alpha - beta - 1) * gamma(alpha + beta + 2) / (gamma(alpha + 1) * gamma(beta + 1)))
      case (1)
        ja = 0.5 * sqrt(2**(-alpha - beta - 1) * gamma(alpha + beta + 2) / (gamma(alpha + 1) * gamma(beta + 1))) * &
             & sqrt((alpha + beta + 3) / (alpha + 1) / (beta + 1)) * ((alpha + beta + 2) * x + alpha - beta)
      case default
        ja = (x * jacobi_ve(x, alpha, beta, n - 1) - &
             & CalcJacobiAn(n - 1, alpha, beta) * jacobi_ve(x, alpha, beta, n - 2) - &
             & CalcJacobiBn(n - 1, alpha, beta) * jacobi_ve(x, alpha, beta, n - 1)) / &
             & CalcJacobiAn(n, alpha, beta)
    end select
  end function jacobi_ve

  function CalcJacobiAn(n, alpha, beta)
    integer, intent(in) :: n
    real(DP), intent(in) :: alpha, beta
    real(DP) CalcJacobiAn

    CalcJacobiAn = 2 / (2 * n + alpha + beta) * &
                   & sqrt(n * (n + alpha + beta) * (n + alpha) * (n + beta) &
                   & / (2 * n + alpha + beta - 1) / (2 * n + alpha + beta + 1))
  end function CalcJacobiAn

  function CalcJacobiBn(n, alpha, beta)
    integer, intent(in) :: n
    real(DP), intent(in) :: alpha, beta
    real(DP) CalcJacobiBn

    CalcJacobiBn = (beta**2 - alpha**2) / (2 * n + alpha + beta) / (2 * n + alpha + beta + 2)
  end function CalcJacobiBn
end module MathFunc
